/* File: draw.cc
 * Copyright (c) 2003 James A. Strother
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <cstdlib>
#include <iostream>
#include <fstream>
#include <string>
#include <queue>
#include <math.h>
#include "gosiv.h"

/* Declare types */
struct draw_vars
{
  double        arrow_scale;
  double        canvas_scale;

  double        xrange_lower;
  double        xrange_upper;
  double        yrange_lower;
  double        yrange_upper;

  int           background_type;
  double        background_stroke_weight;
  gosiv_color   background_stroke_color;
  gosiv_color   background_fill_color;

  int           arrowtail_type;
  double        arrowtail_width;
  double        arrowtail_stroke_weight;
  gosiv_color   arrowtail_stroke_color;
  gosiv_color   arrowtail_fill_color;

  int           arrowhead_type;
  double        arrowhead_width;
  double        arrowhead_height;
  double        arrowhead_angle;
  double        arrowhead_stroke_weight;
  gosiv_color   arrowhead_stroke_color;
  gosiv_color   arrowhead_fill_color;

  int           legend_type;
  int           legend_placement;
  double        legend_horiz_margin;
  double        legend_vert_margin;
  double        legend_horiz_inset;
  double        legend_vert_inset;
  double        legend_stroke_weight;
  gosiv_color   legend_stroke_color;
  gosiv_color   legend_fill_color;

  double        legend_arrow_length;
  string        legend_arrow_label;
  int           legend_arrow_label_font;
  gosiv_color   legend_arrow_label_color;
  double        legend_arrow_label_size;

  int           legend_arrowtail_type;
  double        legend_arrowtail_width;
  double        legend_arrowtail_stroke_weight;
  gosiv_color   legend_arrowtail_stroke_color;
  gosiv_color   legend_arrowtail_fill_color;

  int           legend_arrowhead_type;
  double        legend_arrowhead_width;
  double        legend_arrowhead_height;
  double        legend_arrowhead_angle;
  double        legend_arrowhead_stroke_weight;
  gosiv_color   legend_arrowhead_stroke_color;
  gosiv_color   legend_arrowhead_fill_color;
  
  double        legend_ruler_length;
  string        legend_ruler_label;
  int           legend_ruler_label_font;
  gosiv_color   legend_ruler_label_color;
  double        legend_ruler_label_size;
  double        legend_ruler_width;
  double        legend_ruler_stroke_weight;
  gosiv_color   legend_ruler_stroke_color;
  gosiv_color   legend_ruler_fill_color;

  draw_vars();
};

draw_vars::draw_vars() 
{
}

struct draw_vec
{
  double x;
  double y;
  double u;
  double v;
  
  draw_vec();
};

draw_vec::draw_vec()
{
  x = 0;
  y = 0;
  u = 0;
  v = 0;
}

/* Function prototypes */
void draw_error(const string&);
void draw_trim_space(const string&);
void draw_trim_quotes(const string&);

bool draw_parse_opt0a(const string&,string&);
bool draw_parse_opt0b(const string&,const string&,bool&);
bool draw_parse_opt1a(const string&,string&,string&);
bool draw_parse_opt1b(const string&,const string&,const string&,string&);
void draw_parse_cmd(int,char*[],draw_vars&,string&,string&,string&,string&,bool&);

bool draw_parse_var_type(const string&,const string&,const string&,int&);
bool draw_parse_var_str(const string&,const string&,const string&,string&);
bool draw_parse_var_double(const string&,const string&,const string&,double&);
bool draw_parse_var_color(const string&,const string&,const string&,gosiv_color&);
void draw_parse_var_line(const string&,const string&,draw_vars&);
void draw_parse_vars(istream*,draw_vars&);

void draw_parse_vecs(istream*,queue<draw_vec>&);

void draw_inform(const string&, double);
void draw_guess_var(draw_vars&,queue<draw_vec>&,bool);

void draw_put_back(const draw_vars&,gosiv&);
void draw_put_vecs(const draw_vars&,queue<draw_vec>&,gosiv&);
void draw_put_legend(const draw_vars&,gosiv&);
void draw_put_all(draw_vars&,queue<draw_vec>&,const string&,const string&,bool);

/* Function definitions */
void draw_error(const string& msg)
{
  cerr << "osiv_draw error: " << msg << endl;
  exit(EXIT_FAILURE);
}

void draw_trim_space (string& S)
{
  size_t st = S.find_first_not_of(" \t\n");
  size_t et = S.find_last_not_of(" \t\n");

  if (st == string::npos)
    S.erase();
  else
    S = S.substr(st,et-st+1);
}

void draw_trim_quotes(string& S)
{
  size_t st = S.find_first_of('\"');
  size_t et = S.find_last_of('\"');

  if (st != string::npos)
    S = S.substr(st+1,et-st-1);
}

bool draw_parse_opt0a(const string& in,
		      string& out)
{
  if (in.at(0) == '-' && in != "-")
    return false;

  out = in;
  return true;
}

bool draw_parse_opt0b(const string& in,
		      const string& arg,
		      bool& out)
{
  if (in != arg) return false;

  out = true;
  return true;
}

bool draw_parse_opt1a(const string& in,
		      string& var,
		      string& val)
{
  if (in.length() <= 2 || in.substr(0,2) != "--")
    return false;

  /* Find breakpoints */
  size_t c0 = 1;
  size_t c1 = in.find_first_of('=');
  size_t c2 = in.length();
  if (c1 == string::npos)
    return false;

  /* Make cuts */
  var = in.substr(c0+1,c1-c0-1);
  val = in.substr(c1+1,c2-c1-1);

  return true;
}

bool draw_parse_opt1b(const string& in0,
		      const string& in1,
		      const string& arg,
		      string& out)
{
  if (in0 != arg) return false;

  out = in1;
  return true;
}

void draw_parse_cmd(int argc,
		    char* argv[],
		    draw_vars& v,
		    string& fni,
		    string& fnp,
		    string& fno,
		    string& typ,
		    bool& quiet)
{
  /* Set defaults */
  fni = "-";
  fnp = "";
  fno = "-";
  typ = "eps";
  quiet = false;

  /* Parse command */
  int i = 1;
  while (i < argc)
    {
      /* Parse zero arguments */
      if (draw_parse_opt0a(argv[i],fni) ||
	  draw_parse_opt0b(argv[i],"-q",quiet))
	{
	  i += 1;
	  continue;
	}

      /* Parse parameters */
      string var,val;
      if (draw_parse_opt1a(argv[i],var,val))
	{
	  draw_parse_var_line(var,val,v);
	  i += 1;
	  continue;
	}

      /* Parse options */
      if (argc > i+1 &&
	  (draw_parse_opt1b(argv[i],argv[i+1],"-o",fno) ||
	   draw_parse_opt1b(argv[i],argv[i+1],"-p",fnp) ||
	   draw_parse_opt1b(argv[i],argv[i+1],"-t",typ)))
	{
	  i += 2;
	  continue;
	}

      /* Fail out */
      draw_error("unable to parse option: " + (string)argv[i]);

    }

  /* Check arguments */
  if (typ != "eps" && typ != "svg")
    draw_error("unrecognized output type: " + typ);

}

bool draw_parse_var_type(const string& var,
			 const string& val,
			 const string& arg,
			 int& type)
{
  if (var != arg) return false;

  if (var == "background_type")
    {
      if (val == "solid")
	type = GOSIV_BACK_SOLID;
      else if (val == "none")
	type = GOSIV_BACK_NONE;
      else
	draw_error("bad background type: " + val);
    }
  else if (var == "arrowtail_type" ||
	   var == "legend_arrowtail_type")
    {
      if (val == "line")
	type = GOSIV_ARROWTAIL_LINE;
      else if (val == "block")
	type = GOSIV_ARROWTAIL_BLOCK;
      else if (val == "none")
	type = GOSIV_ARROWTAIL_NONE;
      else
	draw_error("bad arrowtail type: " + val);
    }
  else if (var == "arrowhead_type" ||
	   var == "legend_arrowhead_type")
    {
      if (val == "outline")
	type = GOSIV_ARROWHEAD_OUTLINE;
      else if (val == "lines")
	type = GOSIV_ARROWHEAD_LINES;
      else if (val == "curvy")
	type = GOSIV_ARROWHEAD_CURVY;
      else if (val == "none")
	type = GOSIV_ARROWHEAD_NONE;
      else
	draw_error("bad arrowhead type: " + val);
    }
  else if (var == "legend_type")
    {
      if (val == "solid")
	type = GOSIV_LEGEND_SOLID;
      else if (val == "none")
	type = GOSIV_LEGEND_NONE;
      else
	draw_error("bad legend type " + val);
    }
  else if (var == "legend_placement")
    {
      if (val == "top_left")
	type = GOSIV_LEGEND_TOP_LEFT;
      else if (val == "top_right")
	type = GOSIV_LEGEND_TOP_RIGHT;
      else if (val == "bottom_left")
	type = GOSIV_LEGEND_BOTTOM_LEFT;
      else if (val == "bottom_right")
	type = GOSIV_LEGEND_BOTTOM_RIGHT;
      else
	draw_error("bad legend placement: " + val);
    }
  else if (var == "legend_arrow_label_font" ||
	   var == "legend_ruler_label_font")
    {
      if (val == "courier")
	type = GOSIV_FONT_COURIER;
      else if (val == "times")
	type = GOSIV_FONT_TIMES;
      else if (val == "times_bold")
	type = GOSIV_FONT_TIMES_BOLD;
      else if (val == "times_italic")
	type = GOSIV_FONT_TIMES_ITALIC;
      else if (val == "helvetica")
	type = GOSIV_FONT_HELVETICA;
      else if (val == "helvetica_bold")
	type = GOSIV_FONT_HELVETICA_BOLD;
      else if (val == "helvetica_italic")
	type = GOSIV_FONT_HELVETICA_ITALIC;
      else
	draw_error("bad font: " + val);
    }
  else
    {
      draw_error("unrecognized parameter: " + var);
    }

  return true;
}

bool draw_parse_var_str(const string& var,
			const string& val,
			const string& arg,
			string& str)
{
  if (var != arg) return false;

  str = val;
  draw_trim_quotes(str);

  return true;
}

bool draw_parse_var_double(const string& var,
			   const string& val,
			   const string& arg,
			   double& number)
{
  if (var != arg) return false;;

  char* endptr;
  number = strtod(val.c_str(),&endptr);
  if (*endptr != '\0')
    draw_error("bad double: " + val);

  return true;
}

bool draw_parse_var_color(const string& var,
			  const string& val,
			  const string& arg,
			  gosiv_color& color)
{
  if (var != arg) return false;

  if (val.length() > 5 &&
      val.substr(0,4) == "rgb(" &&
      val.at(val.length()-1) == ')')
    {
      /* Find break points */
      size_t c0 = 3;

      size_t c1 = val.find_first_of(',',c0+1);
      if (c1 == string::npos) 
	draw_error("bad color: " + val);

      size_t c2 = val.find_first_of(',',c1+1);
      if (c2 == string::npos)
	draw_error("bad color: " + val);

      size_t c3 = val.length()-1;

      /* Make cuts */
      string rs = val.substr(c0+1,c1-c0-1);
      string gs = val.substr(c1+1,c2-c1-1);
      string bs = val.substr(c2+1,c3-c2-1);
      draw_trim_space(rs);
      draw_trim_space(gs);
      draw_trim_space(bs);

      if (rs.empty() || gs.empty() || bs.empty())
	draw_error("bad color: " + val);

      /* Convert to numbers */
      char* endptr;
      color.r = strtod(rs.c_str(),&endptr);
      if (*endptr != '\0')
	draw_error("bad color: " + val);

      color.g = strtod(gs.c_str(),&endptr);
      if (*endptr != '\0')
	draw_error("bad color: " + val);

      color.b = strtod(bs.c_str(),&endptr);
      if (*endptr != '\0')
	draw_error("bad color: " + val);
    }
  else if (val == "white")
    color = gosiv_color(1,1,1);
  else if (val == "lightGray")
    color = gosiv_color(0.9,0.9,0.9);
  else if (val == "gray")
    color = gosiv_color(0.75,0.75,0.75);
  else if (val == "darkGray")
    color = gosiv_color(0.5,0.5,0.5);
  else if (val == "black")
    color = gosiv_color(0,0,0);
  else if (val == "yellow")
    color = gosiv_color(1,1,0);
  else if (val == "blue")
    color = gosiv_color(0,0,1);
  else if (val == "red")
    color = gosiv_color(1,0,0);
  else if (val == "green")
    color = gosiv_color(0,1,0);
  else if (val == "purple")
    color = gosiv_color(0.6705,0.5098,1);
  else if (val == "pink")
    color = gosiv_color(.75,0,0);
  else if (val == "brown")
    color = gosiv_color(0.5451,0.1373,0.1373);
  else if (val == "orange")
    color = gosiv_color(1,.6471,0);
  else if (val == "cyan")
    color = gosiv_color(0,1,1);
  else if (val == "magenta")
    color = gosiv_color(1,0,1);
  else if (val == "coffee")
    color = gosiv_color(0.4824,0.2784,0.0784);
  else
    draw_error("bad color: " + val);

  return true;
}

void draw_parse_var_line(const string& var,
			 const string& val,
			 draw_vars& v)
{
  if (!(
	draw_parse_var_double(var,val,"arrow_scale",v.arrow_scale) ||
	draw_parse_var_double(var,val,"canvas_scale",v.canvas_scale) ||
	draw_parse_var_double(var,val,"xrange_lower",v.xrange_lower) ||
	draw_parse_var_double(var,val,"xrange_upper",v.xrange_upper) ||
	draw_parse_var_double(var,val,"yrange_lower",v.yrange_lower) ||
	draw_parse_var_double(var,val,"yrange_upper",v.yrange_upper) ||
	draw_parse_var_type(var,val,"background_type",v.background_type) ||
	draw_parse_var_double(var,val,"background_stroke_weight",v.background_stroke_weight) ||
	draw_parse_var_color(var,val,"background_stroke_color",v.background_stroke_color) ||
	draw_parse_var_color(var,val,"background_fill_color",v.background_fill_color) ||
	draw_parse_var_type(var,val,"arrowtail_type",v.arrowtail_type) ||
	draw_parse_var_double(var,val,"arrowtail_width",v.arrowtail_width) ||
	draw_parse_var_double(var,val,"arrowtail_stroke_weight",v.arrowtail_stroke_weight) ||
	draw_parse_var_color(var,val,"arrowtail_stroke_color",v.arrowtail_stroke_color) ||
	draw_parse_var_color(var,val,"arrowtail_fill_color",v.arrowtail_fill_color) ||
	draw_parse_var_type(var,val,"arrowhead_type",v.arrowhead_type) ||
	draw_parse_var_double(var,val,"arrowhead_width",v.arrowhead_width) ||
	draw_parse_var_double(var,val,"arrowhead_height",v.arrowhead_height) ||
	draw_parse_var_double(var,val,"arrowhead_angle",v.arrowhead_angle) ||
	draw_parse_var_double(var,val,"arrowhead_stroke_weight",v.arrowhead_stroke_weight) ||
	draw_parse_var_color(var,val,"arrowhead_stroke_color",v.arrowhead_stroke_color) ||
	draw_parse_var_color(var,val,"arrowhead_fill_color",v.arrowhead_fill_color) ||
	draw_parse_var_type(var,val,"legend_type",v.legend_type) ||
	draw_parse_var_type(var,val,"legend_placement",v.legend_placement) ||
	draw_parse_var_double(var,val,"legend_horiz_margin",v.legend_horiz_margin) ||
	draw_parse_var_double(var,val,"legend_vert_margin",v.legend_vert_margin) ||
	draw_parse_var_double(var,val,"legend_horiz_inset",v.legend_horiz_inset) ||
	draw_parse_var_double(var,val,"legend_vert_inset",v.legend_vert_inset) ||
	draw_parse_var_double(var,val,"legend_stroke_weight",v.legend_stroke_weight) ||
	draw_parse_var_color(var,val,"legend_stroke_color",v.legend_stroke_color) ||
	draw_parse_var_color(var,val,"legend_fill_color",v.legend_fill_color) ||
	draw_parse_var_double(var,val,"legend_arrow_length",v.legend_arrow_length) ||
	draw_parse_var_str(var,val,"legend_arrow_label",v.legend_arrow_label) ||
	draw_parse_var_type(var,val,"legend_arrow_label_font",v.legend_arrow_label_font) ||
	draw_parse_var_color(var,val,"legend_arrow_label_color",v.legend_arrow_label_color) ||
	draw_parse_var_double(var,val,"legend_arrow_label_size",v.legend_arrow_label_size) ||
	draw_parse_var_type(var,val,"legend_arrowtail_type",v.legend_arrowtail_type) ||
	draw_parse_var_double(var,val,"legend_arrowtail_width",v.legend_arrowtail_width) ||
	draw_parse_var_double(var,val,"legend_arrowtail_stroke_weight",v.legend_arrowtail_stroke_weight) ||
	draw_parse_var_color(var,val,"legend_arrowtail_stroke_color",v.legend_arrowtail_stroke_color) ||
	draw_parse_var_color(var,val,"legend_arrowtail_fill_color",v.legend_arrowtail_fill_color) ||
	draw_parse_var_type(var,val,"legend_arrowhead_type",v.legend_arrowhead_type) ||
	draw_parse_var_double(var,val,"legend_arrowhead_width",v.legend_arrowhead_width) ||
	draw_parse_var_double(var,val,"legend_arrowhead_height",v.legend_arrowhead_height) ||
	draw_parse_var_double(var,val,"legend_arrowhead_angle",v.legend_arrowhead_angle) ||
	draw_parse_var_double(var,val,"legend_arrowhead_stroke_weight",v.legend_arrowhead_stroke_weight) ||
	draw_parse_var_color(var,val,"legend_arrowhead_stroke_color",v.legend_arrowhead_stroke_color) ||
	draw_parse_var_color(var,val,"legend_arrowhead_fill_color",v.legend_arrowhead_fill_color) ||
	draw_parse_var_double(var,val,"legend_ruler_length",v.legend_ruler_length) ||
	draw_parse_var_str(var,val,"legend_ruler_label",v.legend_ruler_label) ||
	draw_parse_var_type(var,val,"legend_ruler_label_font",v.legend_ruler_label_font) ||
	draw_parse_var_color(var,val,"legend_ruler_label_color",v.legend_ruler_label_color) ||
	draw_parse_var_double(var,val,"legend_ruler_label_size",v.legend_ruler_label_size) ||
	draw_parse_var_double(var,val,"legend_ruler_width",v.legend_ruler_width) ||
	draw_parse_var_double(var,val,"legend_ruler_stroke_weight",v.legend_ruler_stroke_weight) ||
	draw_parse_var_color(var,val,"legend_ruler_stroke_color",v.legend_ruler_stroke_color) ||
	draw_parse_var_color(var,val,"legend_ruler_fill_color",v.legend_ruler_fill_color)
	))
    draw_error("unrecognized parameter: " + var);
}

void draw_parse_vars (istream* file,
		      draw_vars& v)
{
  string line;  

  /* Traverse file */
  while (!getline(*file,line).eof())
    {
      /* Skip blank lines */
      draw_trim_space(line);
      if (line.empty())
	continue;
     
      /* Find break points */
      size_t c1 = line.find_first_of('=');
      if (c1 == string::npos) 
	draw_error("bad parameter: " + line);

      /* Make cuts */
      string var,val;
      var = line.substr(0,c1);
      val = line.substr(c1+1,line.length()-c1-1);
      draw_trim_space(var);
      draw_trim_space(val);
      draw_trim_quotes(val);

      if (var.empty())
	draw_error("parameter without lvalue: " + line);

      if (val.empty())
	draw_error("parameter without rvalue: " + line);

      /* Parse line */
      draw_parse_var_line(var,val,v);
    }
}

void draw_default_vars(draw_vars& v)
{
  v.arrow_scale = HUGE_VAL;
  v.canvas_scale = HUGE_VAL;

  v.xrange_lower = HUGE_VAL;
  v.xrange_upper = HUGE_VAL;
  v.yrange_lower = HUGE_VAL;
  v.yrange_upper = HUGE_VAL;

  v.background_type = GOSIV_BACK_SOLID;
  v.background_stroke_weight = 1;
  v.background_stroke_color = gosiv_color(0,0,0);
  v.background_fill_color = gosiv_color(1,1,1);

  v.arrowtail_type = GOSIV_ARROWTAIL_LINE;
  v.arrowtail_width = 0.05;
  v.arrowtail_stroke_weight = .5;
  v.arrowtail_stroke_color = gosiv_color(0,0,0);
  v.arrowtail_fill_color = gosiv_color(0,0,0);

  v.arrowhead_type = GOSIV_ARROWHEAD_LINES;
  v.arrowhead_width = 0.25;
  v.arrowhead_height = 0.20;
  v.arrowhead_angle = 90;
  v.arrowhead_stroke_weight = .5;
  v.arrowhead_stroke_color = gosiv_color(0,0,0);
  v.arrowhead_fill_color = gosiv_color(0,0,0);

  v.legend_type = GOSIV_LEGEND_NONE;
  v.legend_placement = GOSIV_LEGEND_BOTTOM_LEFT;
  v.legend_horiz_margin = 10;
  v.legend_vert_margin = 10;
  v.legend_horiz_inset = 5;
  v.legend_vert_inset = 5;
  v.legend_stroke_weight = 1;
  v.legend_stroke_color = gosiv_color(0,0,0);
  v.legend_fill_color = gosiv_color(1,1,1);

  v.legend_arrow_length = HUGE_VAL;
  v.legend_arrow_label = "1 u";
  v.legend_arrow_label_font = GOSIV_FONT_HELVETICA;
  v.legend_arrow_label_color = gosiv_color(0,0,0);
  v.legend_arrow_label_size = 10;

  v.legend_arrowtail_type = GOSIV_ARROWTAIL_LINE;
  v.legend_arrowtail_width = 0.05;
  v.legend_arrowtail_stroke_weight = .5;
  v.legend_arrowtail_stroke_color = gosiv_color(0,0,0);
  v.legend_arrowtail_fill_color = gosiv_color(0,0,0);

  v.legend_arrowhead_type = GOSIV_ARROWHEAD_LINES;
  v.legend_arrowhead_width = 0.30;
  v.legend_arrowhead_height = 0.25;
  v.legend_arrowhead_angle = 90;
  v.legend_arrowhead_stroke_weight = .5;
  v.legend_arrowhead_stroke_color = gosiv_color(0,0,0);
  v.legend_arrowhead_fill_color = gosiv_color(0,0,0);

  v.legend_ruler_length = HUGE_VAL;
  v.legend_ruler_label = "1 u";
  v.legend_ruler_label_font = GOSIV_FONT_HELVETICA;
  v.legend_ruler_label_color = gosiv_color(0,0,0);
  v.legend_ruler_label_size = 10;
  v.legend_ruler_width = .10;
  v.legend_ruler_stroke_weight = .5;
  v.legend_ruler_stroke_color = gosiv_color(0,0,0);
  v.legend_ruler_fill_color = gosiv_color(0,0,0);
}

void draw_parse_vecs(istream* is,
		     queue<draw_vec>& vq)
{
  string line;
  while (!getline(*is,line).eof())
    {
      draw_trim_space(line);
      if (line.empty())
	continue;

      /* Find breakpoints */
      int c0 = -1;

      size_t c1 = line.find_first_of(',',c0+1);
      if (c1 == string::npos)
	draw_error("unable to parse vectors: " + line);

      size_t c2 = line.find_first_of(';',c1+1);
      if (c2 == string::npos) 
	draw_error("unable to parse vectors: " + line);

      size_t c3 = line.find_first_of(',',c2+1);
      if (c3 == string::npos) 
	draw_error("unable to parse vectors: " + line);

      size_t c4 = line.length();

      /* Make cuts */
      string x_str = line.substr(c0+1,c1-c0-1);
      string y_str = line.substr(c1+1,c2-c1-1);
      string u_str = line.substr(c2+1,c3-c2-1);
      string v_str = line.substr(c3+1,c4-c3-1);
      
      draw_trim_space(x_str);
      draw_trim_space(y_str);
      draw_trim_space(u_str);
      draw_trim_space(v_str);

      /* Convert to numbers */
      draw_vec vec;

      char* endptr;
      vec.x = strtod(x_str.c_str(),&endptr);
      if (*endptr != '\0') 
	draw_error("unable to parse vectors: " + line);

      vec.y = strtod(y_str.c_str(),&endptr);
      if (*endptr != '\0')
	draw_error("unable to parse vectors: " + line);

      vec.u = strtod(u_str.c_str(),&endptr);
      if (*endptr != '\0')
	draw_error("unable to parse vectors: " + line);

      vec.v = strtod(v_str.c_str(),&endptr);
      if (*endptr != '\0')
	draw_error("unable to parse vectors: " + line);

      vq.push(vec);
    }
}

void draw_inform (const string& str,
		  double value)
{
  cerr << "set " << str << " to " << value << endl;
}

void draw_guess_var(draw_vars& vars,
		    queue<draw_vec>& vq,
		    bool quiet)
{
  const double ideal_arrow_length = 10;
  const double ideal_canvas_length = 500;
  const double ideal_canvas_margin = 0.025;

  double uavg = 0;
  double umin = HUGE_VAL;
  double umax = -HUGE_VAL;

  double xavg = 0;
  double xmin = HUGE_VAL;
  double xmax = -HUGE_VAL;

  double yavg = 0;
  double ymin = HUGE_VAL;
  double ymax = -HUGE_VAL;

  int loop = vq.size();
  while(loop-- > 0)
    {
      /* Pull of vector */
      draw_vec v;
      v = vq.front();
      vq.pop();
      vq.push(v);

      /* Skip if not in bounds */
      if ((vars.xrange_lower == HUGE_VAL ||
	   v.x > vars.xrange_lower ) &&
	  (vars.xrange_upper == HUGE_VAL ||
	   v.x < vars.xrange_upper ) &&
	  (vars.yrange_lower == HUGE_VAL ||
	   v.y > vars.yrange_lower ) &&
	  (vars.xrange_upper == HUGE_VAL ||
	   v.y < vars.yrange_upper ))
	{
	  /* Calculate statistics */
	  double umag = sqrt(v.u*v.u+v.v*v.v);
	  uavg += umag;
	  if (umag < umin) umin = umag;
	  if (umag > umax) umax = umag;
      
	  xavg += v.x;
	  if (v.x < xmin) xmin = v.x;
	  if (v.x > xmax) xmax = v.x;

	  yavg += v.y;
	  if (v.y < ymin) ymin = v.y;
	  if (v.y > ymax) ymax = v.y;
	}
    }

  uavg /= vq.size();
  xavg /= vq.size();
  yavg /= vq.size();
  
  /* Set defaults */
  if (xmax-xmin < 1e-8)
    {
      xmin -= xavg;
      xmax += xavg;
    }
  if (ymax-ymin < 1e-8)
    {
      ymin -= yavg;
      ymax += yavg;
    }

  if (vars.arrow_scale == HUGE_VAL)
    {
      vars.arrow_scale = ideal_arrow_length/uavg;
      if (!quiet) draw_inform("arrow_scale",vars.arrow_scale);
    }
  if (vars.xrange_lower == HUGE_VAL)
    {
      vars.xrange_lower = xmin - ideal_canvas_margin*(xmax-xmin);
      if (!quiet) draw_inform("xrange_lower",vars.xrange_lower);
    }
  if (vars.xrange_upper == HUGE_VAL)
    {
      vars.xrange_upper = xmax + ideal_canvas_margin*(xmax-xmin);
      if (!quiet) draw_inform("xrange_upper",vars.xrange_upper);
    }
  if (vars.yrange_lower == HUGE_VAL)
    {
      vars.yrange_lower = ymin - ideal_canvas_margin*(ymax-ymin);
      if (!quiet) draw_inform("yrange_lower",vars.yrange_lower);
    }
  if (vars.yrange_upper == HUGE_VAL)
    {
      vars.yrange_upper = ymax + ideal_canvas_margin*(ymax-ymin);
      if (!quiet) draw_inform("yrange_upper",vars.yrange_upper);
    }
  if (vars.canvas_scale == HUGE_VAL)
    {
      double vala = 1/(vars.xrange_upper-vars.xrange_lower);
      double valb = 1/(vars.yrange_upper-vars.yrange_lower);
      vars.canvas_scale = ideal_canvas_length*(vala<valb?vala:valb);
      if (!quiet) draw_inform("canvas_scale",vars.canvas_scale);
    }
  if (vars.legend_arrow_length == HUGE_VAL)
    {
      vars.legend_arrow_length = uavg;
      if (!quiet) draw_inform("legend_arrow_length",uavg);
    }
  if (vars.legend_ruler_length == HUGE_VAL)
    {
      vars.legend_ruler_length = uavg*vars.arrow_scale/vars.canvas_scale;
      if (!quiet) draw_inform("legend_ruler_length",vars.legend_ruler_length);
    }

}

void draw_put_back(const draw_vars& v,gosiv& g)
{
  gosiv_back back;
  back.type           = v.background_type;
  back.stroke_weight  = v.background_stroke_weight;
  back.stroke_color   = v.background_stroke_color;
  back.fill_color     = v.background_fill_color;
  g.drawBack(back);
}

void draw_put_vecs(const draw_vars& vars,
		   queue<draw_vec>& vq,
		   gosiv& g)
{
  /* Construct arrow type */
  gosiv_arrow a;
  a.scale               = vars.arrow_scale;
  a.tail_type           = vars.arrowtail_type;
  a.tail_width          = vars.arrowtail_width;
  a.tail_stroke_weight  = vars.arrowtail_stroke_weight;
  a.tail_stroke_color   = vars.arrowtail_stroke_color;
  a.tail_fill_color     = vars.arrowtail_fill_color;
  a.head_type           = vars.arrowhead_type;
  a.head_width          = vars.arrowhead_width;
  a.head_height         = vars.arrowhead_height;
  a.head_angle          = vars.arrowhead_angle;
  a.head_stroke_weight  = vars.arrowhead_stroke_weight;
  a.head_stroke_color   = vars.arrowhead_stroke_color;
  a.head_fill_color     = vars.arrowhead_fill_color;

  while (!vq.empty())
    {
      draw_vec vec;
      vec = vq.front();
      vq.pop();
      a.x = vec.x;
      a.y = vec.y;
      a.u = vec.u;
      a.v = vec.v;
      g.drawArrow(a);
    }
}

void draw_put_legend(const draw_vars& v,
		     gosiv& g)
{
  gosiv_legend legend;

  legend.type                      = v.legend_type;
  legend.placement                 = v.legend_placement;
  legend.horiz_margin              = v.legend_horiz_margin;
  legend.vert_margin               = v.legend_vert_margin;
  legend.horiz_inset               = v.legend_horiz_inset;
  legend.vert_inset                = v.legend_vert_inset;
  legend.stroke_weight             = v.legend_stroke_weight;
  legend.stroke_color              = v.legend_stroke_color;
  legend.fill_color                = v.legend_fill_color;

  legend.arrow_length              = v.legend_arrow_length;
  legend.arrow_label               = v.legend_arrow_label;
  legend.arrow_label_font          = v.legend_arrow_label_font;
  legend.arrow_label_color         = v.legend_arrow_label_color;
  legend.arrow_label_size          = v.legend_arrow_label_size;

  legend.arrow.scale               = v.arrow_scale;
  legend.arrow.tail_type           = v.legend_arrowtail_type;
  legend.arrow.tail_width          = v.legend_arrowtail_width;
  legend.arrow.tail_stroke_weight  = v.legend_arrowtail_stroke_weight;
  legend.arrow.tail_stroke_color   = v.legend_arrowtail_stroke_color;
  legend.arrow.tail_fill_color     = v.legend_arrowtail_fill_color;
  legend.arrow.head_type           = v.legend_arrowhead_type;
  legend.arrow.head_width          = v.legend_arrowhead_width;
  legend.arrow.head_height         = v.legend_arrowhead_height;
  legend.arrow.head_angle          = v.legend_arrowhead_angle;
  legend.arrow.head_stroke_weight  = v.legend_arrowhead_stroke_weight;
  legend.arrow.head_stroke_color   = v.legend_arrowhead_stroke_color;
  legend.arrow.head_fill_color     = v.legend_arrowhead_fill_color;

  legend.ruler_length              = v.legend_ruler_length;
  legend.ruler_width               = v.legend_ruler_width;
  legend.ruler_stroke_weight       = v.legend_ruler_stroke_weight;
  legend.ruler_stroke_color        = v.legend_ruler_stroke_color;
  legend.ruler_fill_color          = v.legend_ruler_fill_color;
  legend.ruler_label               = v.legend_ruler_label;
  legend.ruler_label_font          = v.legend_ruler_label_font;
  legend.ruler_label_color         = v.legend_ruler_label_color;
  legend.ruler_label_size          = v.legend_ruler_label_size;

  g.drawLegend(legend);
}

void draw_put_all(draw_vars& vars,
		  queue<draw_vec>& vq,
		  const string& ftp,
		  const string& fn,
		  bool quiet)
{
  /* Guess defaults */
  draw_guess_var(vars,vq,quiet);

  /* Put together canvas */
  gosiv_canvas canvas;
  canvas.scale = vars.canvas_scale;
  canvas.xlower = vars.xrange_lower;
  canvas.xupper = vars.xrange_upper;
  canvas.ylower = vars.yrange_lower;
  canvas.yupper = vars.yrange_upper;
  canvas.width = (canvas.xupper-canvas.xlower)*canvas.scale;
  canvas.height = (canvas.yupper-canvas.ylower)*canvas.scale;

  /* Open gosiv */
  gosiv g;
  
  if (ftp == "eps")
    g.openPS(canvas,fn);
  else if (ftp == "svg")
    g.openSVG(canvas,fn);
  else
    draw_error("unrecognized output type: " + ftp);

  /* Write everything */
  draw_put_back(vars,g);
  draw_put_vecs(vars,vq,g);
  draw_put_legend(vars,g);

  /* Close gosiv */
  g.close();
}

int main (int argc, char* argv[])
{
  draw_vars vars;
  queue<draw_vec> vq;
  string fni,fnp;
  string ftp,fno;
  bool quiet;

  /* Set defaults */
  draw_default_vars(vars);

  /* Parse command */
  draw_parse_cmd(argc,argv,vars,fni,
		 fnp,fno,ftp,quiet);

  /* Parse variables */
  if (!fnp.empty())
    {
      ifstream fptr(fnp.c_str());
      if(fptr == NULL)
	draw_error("unable to open parameter file: " + fnp);
      draw_parse_vars(&fptr,vars);
    }

  /* Parse vectors */
  if (fni != "-")
    {
      ifstream fptr(fni.c_str());
      if (fptr == NULL)
	draw_error("unable to open vector file: " + fni);
      draw_parse_vecs(&fptr,vq);
    }
  else
    {
      draw_parse_vecs(&cin,vq);
    }
  if (vq.empty())
    draw_error("empty vector file");

  /* Write output */
  draw_put_all(vars,vq,ftp,fno,quiet);
  
  return EXIT_SUCCESS;
}
