/* File: getargs.c

* Copyright (c) 2002 James A. Strother
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
*/

#include "osiv.h"

static const char*  osiv_key_convert[] =
  {
    "Alg",           /* 00 */
    "PreProc",       /* 01 */
    "Interp",        /* 02 */
    "XSize",         /* 03 */
    "YSize",         /* 04 */
    "XStart",        /* 05 */
    "YStart",        /* 06 */
    "XOffset",       /* 07 */
    "YOffset",       /* 08 */
    "XVectSpace",    /* 09 */
    "YVectSpace",    /* 10 */
    "XNumVects",     /* 11 */
    "YNumVects",     /* 12 */
    "XWinSize",      /* 13 */
    "YWinSize",      /* 14 */
    "XMaxDisp",      /* 15 */
    "YMaxDisp",      /* 16 */
    "XWinLgSize",    /* 17 */
    "YWinLgSize",    /* 18 */
    "FrmStart",      /* 19 */
    "FrmFinish",     /* 20 */
    "FrmSkip",       /* 21 */
    "FrmSet",        /* 22 */
    "NumIter",       /* 23 */
    "WriteMap",      /* 24 */
    "MovieFN",       /* 25 */
    "VectorsFN"      /* 26 */
  };

static void getargs_error(const char* m)
{
  fprintf(stderr, "%s error: %s\n","osiv_getargs",m);
  exit(EXIT_FAILURE);
}

static void getargs_error2(const char* m1,const char* m2)
{
  fprintf(stderr,"%s error: %s%s\n","osiv_getargs",m1,m2);
  exit(EXIT_FAILURE);
}

static void getargs_puint(char* arg,
			  long* number)
{
  char* t;
  *number = strtol(arg,&t,10);
  if (*t != '\0')
    getargs_error2("unable to parse number: ",arg);
}

static void getargs_pustr(char* arg,
			  char* str)
{
  if (strlen(arg) > OSIV_STR_LEN)
    getargs_error2("string too long: ",arg);

  strcpy(str,arg);
}

static void getargs_parse(int argc,
			  char** argv,
			  char* fni,
			  char* fno)
{
  int i;

  /* Go to first real argument */
  argc--;
  argv++;

  /* Set null argument states */
  *fni = '\0';

  /* Set defaults */
  strcpy(fno,"-");

  /* Interpret arguments */
  i = 0;
  while (i < argc)
    {
      char* c = *(argv+i++);
      
      /* Parse filename */
      if (*fni == '\0' && (*c != '-' || !strcmp(c,"-")))
	{
	  if (strlen(c) > OSIV_STR_LEN)
	    getargs_error2("string too long: ",c);

	  strcpy(fni,c);
	  continue;
	}

      /* Parse options with one argument */
      if (i < argc && *c == '-' && strlen(c)==2)
	{
	  switch (*(c+1))
	    {
	    case 'o':
	      getargs_pustr(*(argv+i++),fno);	  
	      break;
	    default:
	      getargs_error2("unrecognized option: ",c);
	      break;
	    }
	  continue;
	}

      /* Fail out */
      getargs_error2("unable to parse: ", c);
    }

  /* Verify argments */
  if (*fni == '\0')
    getargs_error("missing input file argument");
}

static char* getargs_convert (int key)
{
  char* t = (char*)malloc(OSIV_STR_LEN+1);
  if (!t) getargs_error("unable to allocate memory");

  if (key >= 0 && key <= OSIV_KEY_MAX)
    strcpy(t,osiv_key_convert[key]);
  else
    sprintf(t,"%d",key);

  return t;
}

static void getargs_int (osiv_file* fi,
			 FILE* fo)
{
  char *vnm;
  osiv_ioint key,val;
  osiv_iotag ctag;

  /* Read in value key and value */
  if (OsivReadArgInt(fi,&key,&val) ||
      OsivReadTag(fi,&ctag) ||
      OsivTagCmp(ctag,OSIV_GAT_ARG_INT))
    getargs_error("unable to parse file");

  /* Convert key to real */
  vnm = getargs_convert(key);

  /* Print out result */
  if (val == OSIV_IOINT_BLANK)
    fprintf(fo,"%s = blank\n",vnm);
  else
    fprintf(fo,"%s = %d\n",vnm,val);

  /* Free key */
  free(vnm);
}

static void getargs_str (osiv_file* fi,
			 FILE* fo)
{
  char *val,*vnm;
  osiv_ioint key;
  osiv_iotag ctag;

  if (OsivReadArgStr(fi,&key,&val) ||
      OsivReadTag(fi,&ctag) ||
      OsivTagCmp(ctag,OSIV_GAT_ARG_STR))
    getargs_error("unable to parse file");

  /* Convert key */
  vnm = getargs_convert(key);

  /* Print out result */
  fprintf(fo,"%s = %s\n",vnm,val);

  /* Free key */
  free(val);
  free(vnm);
}
    

static void getargs(osiv_file* fi, FILE* fo)
{
  osiv_iotag ctag;

  /* Scan to header */
  if (OsivScanToTag(fi,OSIV_TAG_HEADER) ||
      OsivReadTag(fi,&ctag))
    getargs_error("unable to parse file");

  /* Proceed to thru header */
  while(OsivTagCmp(ctag,OSIV_GAT_HEADER))
    {
      switch (ctag)
	{
	case OSIV_TAG_ARG_INT:
	  getargs_int(fi,fo);
	  break;
	case OSIV_TAG_ARG_STR:
	  getargs_str(fi,fo);
	  break;
	default:
	  getargs_error("unrecognized header tag");
	  break;
	}
      if (OsivReadTag(fi,&ctag))
	getargs_error("unable to parse file");
    }

}

int main (int argc, char** argv)
{
  char *fni,*fno;
  osiv_file* fi;
  FILE* fo;
 
  /* Allocate memory for filenames */
  fni = (char*)malloc(OSIV_STR_LEN+1);
  fno = (char*)malloc(OSIV_STR_LEN+1);
  if (!fni || !fno)
    getargs_error("unable to allocate memory");

  /* Parse incoming data */
  getargs_parse(argc,argv,fni,fno);

  /* Open input file */
  if (!strcmp(fni,"-"))
    fi = (osiv_file*)OsivOpenIO(stdin);
  else
    fi = (osiv_file*)OsivOpen(fni);

  if (!fi)
    getargs_error("unable to open input file");

  /* Open output file */
  if (!strcmp(fno,"-"))
    fo = stdout;
  else
    fo = fopen(fno,"w");

  if (!fo)
    getargs_error("unable to open output file");

  /* Print arguments */
  getargs(fi,fo);

  /* Close file */
  fclose(fo);
  OsivClose(fi);

  /* Clean up */
  free(fno);
  free(fni);

  /* Return */
  exit(EXIT_SUCCESS);
}
