/* File: getmap.c

* Copyright (c) 2002 James A. Strother
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
*/

#include "osiv.h"

struct getmap_box_t
{
  char* fmt;
  osiv_ioreal* ptr;
  struct getmap_box_t* next;
};

typedef struct getmap_box_t getmap_box;

static void getmap_error(char* m)
{
  fprintf(stderr, "%s error: %s\n","osiv_getmap",m);
  exit(EXIT_FAILURE);
}

static void getmap_error2(const char* m1,const char* m2)
{
  fprintf(stderr,"%s error: %s%s\n","osiv_getmap",m1,m2);
  exit(EXIT_FAILURE);
}

static void getmap_puint(char* arg,
			 long* number)
{
  char* t;
  *number = strtol(arg,&t,10);
  if (*t != '\0')
    getmap_error2("unable to parse number: ",arg);
}

static void getmap_pustr(char* arg,
			 char* str)
{
  if (strlen(arg) > OSIV_STR_LEN)
    getmap_error2("string too long: ",arg);

  strcpy(str,arg);
}

static void getmap_unescape (char* str)
{
  int i,nume = 8;
  char usc[] = {'\f','\n','\r','\t','\v','\'','\"','\?'};
  char* esc[] = {"\\f","\\n","\\r","\\t","\\v","\\\'","\\\"","\\\?"};

  for (i=0;i<nume;i++)
    {
      char *c;
      while ((c=strstr(str,esc[i])) != NULL)
	{
	  char *src,*dest;

	  dest = c+1;
	  src = c+strlen(esc[i]);
	  memmove(dest,src,strlen(src)+1);
	
	  *c = usc[i];
	}
    }
}

static void getmap_parse(int argc,
			 char** argv,
			 char* fni,
			 char* fno,
			 long* frm,
			 long* wdw,
			 char* fmt)
{
  int i;

  /* Go to first real argument */
  argc--;
  argv++;

  /* Set null argument states */
  *fni = '\0';
  *frm = -1;
  *wdw = -1;

  /* Set defaults */
  strcpy(fno,"-");
  strcpy(fmt,"%.0u %.0v %+.4e\n");

  /* Traverse arguments */
  i = 0;
  while (i < argc)
    {
      char* c = *(argv+i++);
      
      /* Parse filename */
      if (*fni == '\0' && (*c != '-' || !strcmp(c,"-")))
	{
	  if (strlen(c) > OSIV_STR_LEN)
	    getmap_error2("string too long: ",c);

	  strcpy(fni,c);
	  continue;
	}

      /* Parse options with one argument */
      if (i < argc && *c == '-' && strlen(c)==2)
	{
	  switch (*(c+1))
	    {
	    case 'o':
	      getmap_pustr(*(argv+i++),fno);	  
	      break;
	    case 'p':
	      getmap_pustr(*(argv+i++),fmt);
	      getmap_unescape(fmt);
	      break;
	    case 'f':
	      getmap_puint(*(argv+i++),frm);
	      break;
	    case 'w':
	      getmap_puint(*(argv+i++),wdw);
	      break;
	    default:
	      getmap_error2("unrecognized option: ",c);
	      break;
	    }
	  continue;
	}

      /* Fail out */
      getmap_error2("unable to parse: ", c);
    }

  /* Verify argments */
  if (*fni == '\0')
    getmap_error("missing input file argument");
  if (*frm == -1)
    getmap_error("missing frame number argument");
  if (*wdw == -1)
    getmap_error("missing window number argument");

}

static void getmap_getdisp(osiv_file* fi,
			   osiv_intf* len,
			   osiv_ioreal** disp)
{
  osiv_ioint xdisp,ydisp;

  /* Collect grid info */
  {
    long mark;
    if (OsivScanToTag(fi,OSIV_TAG_HEADER) ||
	OsivTell(fi,&mark) ||
	OsivGrabArgInt(fi,OSIV_KEY_XMAXDISP,&xdisp) ||
	OsivSeek(fi,mark,SEEK_SET) ||
	OsivGrabArgInt(fi,OSIV_KEY_YMAXDISP,&ydisp) ||
	OsivRewind(fi))
      getmap_error("unable to parse data file");
  }

  /* Allocate memory */
  *len = (2*xdisp+1)*(2*ydisp+1);
  *disp = (osiv_ioreal*)malloc(*len*sizeof(osiv_ioreal));
  if (!*disp) getmap_error("unable to allocate memory");

  /* Make grid */
  {
    osiv_ioreal* run = *disp;

    osiv_intf i,j;
    for (i = -ydisp; i <= ydisp; i++)
      {
	for (j = -xdisp; j <= xdisp; j++)
	  {
	    *run++ = j;
	    *run++ = i;
	  }
      }	
  }

}

static void getmap_getcorr(osiv_file* fi,
			   osiv_intm frm,
			   osiv_intf wdw,
			   osiv_intf* len,
			   osiv_ioreal** corr)
{
  if (OsivScanToTag(fi,OSIV_TAG_BODY) ||
      OsivGrabMap(fi,frm,wdw,len,corr) ||
      OsivRewind(fi))
    getmap_error("unable to parse data file");
}

static getmap_box* getmap_getfmt(const char* fmt,
				 osiv_ioreal* uptr,
				 osiv_ioreal* vptr,
				 osiv_ioreal* eptr)
{
  static char* flags = "#0+- ";
  static char* digits = "0123456789";

  const char *c0,*c1;
  getmap_box* head = NULL;
  getmap_box* tail = NULL;

  c1 = c0 = fmt;
  while (strlen(c0) > 0)
    {
      getmap_box* box;

      /* Allocate memory for element */
      box = (getmap_box*)malloc(sizeof(getmap_box));
      if (!box) getmap_error("unable to allocate memory");
      if (!head) head = tail = box;

      /* Find format character */
      c1 = strchr(c0,'%');

      /* Skip to end of format */
      if (c1 == NULL)
	{
	  c1 = c0+strlen(c0)-1;

	  /* Push pointer */
	  box->ptr = NULL;

	  /* Push string */
	  {
	    size_t len = c1-c0+1;

	    box->fmt = (char*)malloc((len+1)*sizeof(char));
	    if (!(box->fmt)) getmap_error("unable to allocate memory");

	    strncpy(box->fmt,c0,len);
	    *(box->fmt+len) = '\0';
	  }
	}
      else
	{
	  /* After `%' */
	  c1++;

	  /* Skip escaped format */
	  if (*c1 == '%')
	    {
	      c1++;
	      continue;
	    }

	  /* Skip over format string */
	  c1 += strspn(c1,flags);
	  c1 += strspn(c1,digits);
	  if (*c1 == '.')
	    c1 += strspn(++c1,digits);

	  /* Push pointer */
	  switch(*c1)
	    {
	    case 'u':
	      box->ptr = uptr;
	      break;
	    case 'v':
	      box->ptr = vptr;
	      break;
	    case 'e':
	      box->ptr = eptr;
	      break;
	    default:
	      getmap_error("bad format string");
	      break;
	    }

	  /* Push string */
	  {
	    size_t len = c1-c0+1;

	    box->fmt = (char*)malloc((len+1)*sizeof(char));
	    if (!(box->fmt)) getmap_error("unable to allocate memory");

	    strncpy(box->fmt,c0,len);
	    *(box->fmt+len-1) = 'f';
	    *(box->fmt+len) = '\0';
	  }
	}

      /* Push box */
      tail->next = box;
      box->next = NULL;
      tail = box;
      
      /* Goto next */
      c0 = c1+1;
    }

  return head;
}

static void getmap_print(FILE* file,
			 const char* fmt,
			 osiv_intf len,
			 osiv_ioreal* disp,
			 osiv_ioreal* corr)
{
  osiv_intf i;
  osiv_ioreal u,v,e;
  getmap_box* list;

  /* Convert format string */
  list = getmap_getfmt(fmt,&u,&v,&e);

  /* Traverse file */
  for (i = 0; i < len/2; i++)
    {
      getmap_box* tmp;

      u = *disp++;
      v = *disp++;
      e = *corr++;

      tmp = list;
      while (tmp)
	{
	  if (tmp->ptr)
	    fprintf(file,tmp->fmt,*(tmp->ptr));
	  else
	    fprintf(file,tmp->fmt);
	}
    }

  /* Cleanup format string */
  while(list)
    {
      getmap_box* tmp;
      tmp = list->next;
      free(list->fmt);
      free(list);
      list = tmp;
    }  
}

static void getmap(osiv_file* fi,
		   FILE* fo,
		   osiv_intm frm,
		   osiv_intf wind,
		   char* fmt)
{
  osiv_ioreal* disp;
  osiv_ioreal* corr;
  osiv_intf dlen,clen;

  /* Read in data */
  getmap_getdisp(fi,&dlen,&disp);
  getmap_getcorr(fi,frm,wind,&clen,&corr);
  if (dlen != clen)
    getmap_error("unable to parse data file");

  /* Print out vectors */
  getmap_print(fo,fmt,dlen,disp,corr);

  /* Clean up */
  free(corr);
  free(disp);
}

int main (int argc, char** argv)
{
  char *fni,*fno;
  osiv_file* fi;
  FILE* fo;
  long frm,wdw;
  char* fmt;

  /* Allocate memory for filenames */
  fni = (char*)malloc(OSIV_STR_LEN+1);
  fno = (char*)malloc(OSIV_STR_LEN+1);
  fmt = (char*)malloc(OSIV_STR_LEN+1);
  if (!fni || !fno | !fmt)
    getmap_error("unable to allocate memory");

  /* Parse incoming data */
  getmap_parse(argc,argv,fni,fno,&frm,&wdw,fmt);

  /* Open input file */
  if (!strcmp(fni,"-"))
    fi = (osiv_file*)OsivOpenIO(stdin);
  else
    fi = (osiv_file*)OsivOpen(fni);

  if (!fi)
    getmap_error("unable to open input file");

  /* Open output file */
  if (!strcmp(fno,"-"))
    fo = stdout;
  else
    fo = fopen(fno,"w");

  if (!fo)
    getmap_error("unable to open output file");

  /* Print out data */
  getmap(fi,fo,frm,wdw,fmt);

  /* Close file */
  fclose(fo);
  OsivClose(fi);

  /* Clean up */
  free(fno);
  free(fni);

  /* Return */
  exit(EXIT_SUCCESS);

}
