/* File: getvects.c

* Copyright (c) 2002 James A. Strother
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
*/

#include "osiv.h"

struct getvects_box_t
{
  char* fmt;
  osiv_ioreal* ptr;
  struct getvects_box_t* next;
};

typedef struct getvects_box_t getvects_box;

static void getvects_error(char* m)
{
  fprintf(stderr, "%s error: %s\n","osiv_getvects",m);
  exit(EXIT_FAILURE);
}

static void getvects_error2(const char* m1,const char* m2)
{
  fprintf(stderr,"%s error: %s%s\n","osiv_getvects",m1,m2);
  exit(EXIT_FAILURE);
}

static void getvects_puint(char* arg,
			   long* number)
{
  char* t;
  *number = strtol(arg,&t,10);
  if (*t != '\0')
    getvects_error2("unable to parse number: ",arg);
}

static void getvects_pustr(char* arg,
			   char* str)
{
  if (strlen(arg) > OSIV_STR_LEN)
    getvects_error2("string too long: ",arg);

  strcpy(str,arg);
}

static void getvects_unescape (char* str)
{
  int i,nume = 8;
  char usc[] = {'\f','\n','\r','\t','\v','\'','\"','\?'};
  char* esc[] = {"\\f","\\n","\\r","\\t","\\v","\\\'","\\\"","\\\?"};

  for (i=0;i<nume;i++)
    {
      char *c;
      while ((c=strstr(str,esc[i])) != NULL)
	{
	  char *src,*dest;

	  dest = c+1;
	  src = c+strlen(esc[i]);
	  memmove(dest,src,strlen(src)+1);
	
	  *c = usc[i];
	}
    }
}

static void getvects_parse(int argc,
			   char** argv,
			   char* fni,
			   char* fno,
			   long* frm,
			   char* fmt)
{
  int i;

  /* Go to first real argument */
  argc--;
  argv++;

  /* Set null argument states */
  *fni = '\0';
  *frm = -1;

  /* Set defaults */
  strcpy(fno,"-");
  strcpy(fmt,"%.0x, %.0y; %+.8u, %+.8v\n");

  /* Traverse arguments */
  i = 0;
  while (i < argc)
    {
      char* c = *(argv+i++);
      
      /* Parse filename */
      if (*fni == '\0' && (*c != '-' || !strcmp(c,"-")))
	{
	  if (strlen(c) > OSIV_STR_LEN)
	    getvects_error2("string too long: ",c);

	  strcpy(fni,c);
	  continue;
	}

      /* Parse options with one argument */
      if (i < argc && *c == '-' && strlen(c)==2)
	{
	  switch (*(c+1))
	    {
	    case 'o':
	      getvects_pustr(*(argv+i++),fno);	  
	      break;
	    case 'p':
	      getvects_pustr(*(argv+i++),fmt);
	      getvects_unescape(fmt);
	      break;
	    case 'f':
	      getvects_puint(*(argv+i++),frm);
	      break;
	    default:
	      getvects_error2("unrecognized option: ",c);
	      break;
	    }
	  continue;
	}

      /* Fail out */
      getvects_error2("unable to parse: ", c);
    }

  /* Verify argments */
  if (*fni == '\0')
    getvects_error("missing input file argument");
  if (*frm == -1)
    getvects_error("missing frame number argument");

}

static void getvects_getpos(osiv_file* fi,
			    osiv_intf* len,
			    osiv_ioreal** pos)
{
  osiv_ioint xnum,ynum;
  osiv_ioint xstart,ystart;
  osiv_ioint xspace,yspace;

  /* Collect grid info */
  {
    long mark;
    if (OsivScanToTag(fi,OSIV_TAG_HEADER) ||
	OsivTell(fi,&mark) ||
	OsivGrabArgInt(fi,OSIV_KEY_XSTART,&xstart) ||
	OsivSeek(fi,mark,SEEK_SET) ||
	OsivGrabArgInt(fi,OSIV_KEY_YSTART,&ystart) ||
	OsivSeek(fi,mark,SEEK_SET) ||
	OsivGrabArgInt(fi,OSIV_KEY_XNUMVECTS,&xnum) ||
	OsivSeek(fi,mark,SEEK_SET) ||
	OsivGrabArgInt(fi,OSIV_KEY_YNUMVECTS,&ynum) ||
	OsivSeek(fi,mark,SEEK_SET) ||
	OsivGrabArgInt(fi,OSIV_KEY_XVECTSPACE,&xspace) ||
	OsivSeek(fi,mark,SEEK_SET) ||
	OsivGrabArgInt(fi,OSIV_KEY_YVECTSPACE,&yspace) ||
	OsivRewind(fi))
      getvects_error("unable to parse data file");
  }

  /* Allocate memory */
  *len = 2*xnum*ynum;
  *pos = (osiv_ioreal*)malloc(*len*sizeof(osiv_ioreal));
  if (!*pos) getvects_error("unable to allocate memory");

  /* Make grid */
  {
    osiv_ioreal* run = *pos;

    osiv_intf i,j;
    for (i = 0; i < ynum; i++)
      {
	for (j = 0; j < xnum; j++)
	  {
	    *run++ = xstart+xspace*j;
	    *run++ = ystart+yspace*i;
	  }
      }	
  }
}

static void getvects_getdisp(osiv_file* fi,
			     osiv_intm frm,
			     osiv_intf* len,
			     osiv_ioreal** disp)
{
  if (OsivScanToTag(fi,OSIV_TAG_BODY) ||
      OsivGrabVectors(fi,frm,len,disp) ||
      OsivRewind(fi))
    getvects_error("unable to parse data file");
}

static getvects_box* getvects_getfmt(const char* fmt,
				     osiv_ioreal* xptr,
				     osiv_ioreal* yptr,
				     osiv_ioreal* uptr,
				     osiv_ioreal* vptr)
{
  static char* flags = "#0+- ";
  static char* digits = "0123456789";

  const char *c0,*c1;
  getvects_box* head = NULL;
  getvects_box* tail = NULL;

  c1 = c0 = fmt;
  while (strlen(c0) > 0)
    {
      getvects_box* box;

      /* Allocate memory for element */
      box = (getvects_box*)malloc(sizeof(getvects_box));
      if (!box) getvects_error("unable to allocate memory");
      if (!head) head = tail = box;

      /* Find format character */
      c1 = strchr(c0,'%');

      /* Skip to end of format */
      if (c1 == NULL)
	{
	  c1 = c0+strlen(c0)-1;

	  /* Push pointer */
	  box->ptr = NULL;

	  /* Push string */
	  {
	    size_t len = c1-c0+1;

	    box->fmt = (char*)malloc((len+1)*sizeof(char));
	    if (!(box->fmt))
	      getvects_error("unable to allocate memory");

	    strncpy(box->fmt,c0,len);
	    *(box->fmt+len) = '\0';
	  }

	}
      else
	{
	  /* After `%' */
	  c1++;

	  /* Skip escaped format */
	  if (*c1 == '%')
	    {
	      c1++;
	      continue;
	    }

	  /* Skip over format string */
	  c1 += strspn(c1,flags);
	  c1 += strspn(c1,digits);
	  if (*c1 == '.')
	    c1 += strspn(++c1,digits);

	  /* Push pointer */
	  switch(*c1)
	    {
	    case 'x':
	      box->ptr = xptr;
	      break;
	    case 'y':
	      box->ptr = yptr;
	      break;
	    case 'u':
	      box->ptr = uptr;
	      break;
	    case 'v':
	      box->ptr = vptr;
	      break;
	    default:
	      getvects_error("bad format string");
	      break;
	    }

	  /* Push string */
	  {
	    size_t len = c1-c0+1;

	    box->fmt = (char*)malloc((len+1)*sizeof(char));
	    if (!(box->fmt))
	      getvects_error("unable to allocate memory");

	    strncpy(box->fmt,c0,len);
	    *(box->fmt+len-1) = 'f';
	    *(box->fmt+len) = '\0';
	  }
	}

      /* Push box */
      tail->next = box;
      box->next = NULL;
      tail = box;
      
      /* Goto next */
      c0 = c1+1;
    }
  
  return head;
}

static void getvects_print(FILE* file,
			   const char* fmt,
			   osiv_intf len,
			   osiv_ioreal* pos,
			   osiv_ioreal* disp)
{
  osiv_intf i;
  osiv_ioreal x,y,u,v;
  getvects_box* list;

  /* Convert format string */
  list = getvects_getfmt(fmt,&x,&y,&u,&v);

  /* Traverse file */
  for (i = 0; i < len/2; i++)
    {
      getvects_box* tmp;

      x = *pos++;
      y = *pos++;
      u = *disp++;
      v = *disp++;

      tmp = list;
      while (tmp)
	{
	  if (tmp->ptr)
	    fprintf(file,tmp->fmt,*(tmp->ptr));
	  else
	    fprintf(file,tmp->fmt);
	  tmp = tmp->next;
	}
    }

  /* Cleanup format string */
  while(list)
    {
      getvects_box* tmp;
      tmp = list->next;
      free(list->fmt);
      free(list);
      list = tmp;
    }  
}

static void getvects (osiv_file* fi,
		      FILE* fo,
		      long frm,
		      char* fmt)
{
  osiv_ioreal* pos;
  osiv_ioreal* disp;
  osiv_intf plen,dlen;

  /* Read in data */
  getvects_getpos(fi,&plen,&pos);
  getvects_getdisp(fi,frm,&dlen,&disp);

  if (plen != dlen)
    getvects_error("unable to parse data file");

  /* Print out vectors */
  getvects_print(fo,fmt,plen,pos,disp);

  /* Clean up */
  free(pos);
  free(disp);
}

int main (int argc, char** argv)
{
  long frm;
  char* fni;
  char* fno;
  char* fmt;
  osiv_file* fi;
  FILE* fo;

  /* Allocate memory for strings */
  fni = (char*)malloc(OSIV_STR_LEN+1);
  fno = (char*)malloc(OSIV_STR_LEN+1);
  fmt = (char*)malloc(OSIV_STR_LEN+1);
  if (!fni || !fno || !fmt)
    getvects_error("unable to allocate memory");

  /* Parse incoming data */
  getvects_parse(argc,argv,fni,fno,&frm,fmt);

  /* Open input file */
  if (!strcmp(fni,"-"))
    fi = (osiv_file*)OsivOpenIO(stdin);
  else
    fi = (osiv_file*)OsivOpen(fni);

  if (!fi)
    getvects_error("unable to open input file");

  /* Open output file */
  if (!strcmp(fno,"-"))
    fo = stdout;
  else
    fo = fopen(fno,"w");

  if (!fo)
    getvects_error("unable to open output file");

  /* Print out data */
  getvects(fi,fo,frm,fmt);

  /* Close file */
  OsivClose(fi);
  fclose(fo);

  /* Clean up */
  free(fno);
  free(fni);
  free(fmt);

  /* Return */
  exit(EXIT_SUCCESS);
}
